-- First source of threat used consumes misdirection. Change to false when Blizzard fixes this.
local USE_BUGGED_MISDIRECTION = false

local _G = _G
local assert = _G.assert
local tonumber = _G.tonumber
local ipairs = _G.ipairs
local pairs = _G.pairs
local tostring = _G.tostring
local tinsert = _G.tinsert
local tremove = _G.tremove
local max = _G.max
local min = _G.min
local time = _G.time

local GetInventoryItemLink = _G.GetInventoryItemLink
local GetItemGem = _G.GetItemGem
local UnitMana = _G.UnitMana
local UnitManaMax = _G.UnitManaMax
local UnitHealth = _G.UnitHealth
local UnitHealthMax = _G.UnitHealthMax
local UnitName = _G.UnitName
local UnitClass = _G.UnitClass
local UnitIsDead = _G.UnitIsDead

local ParserLib = _G.ParserLib
local ParserLib_SELF = _G.ParserLib_SELF

local AceLibrary = _G.AceLibrary

local MAJOR_VERSION = "Threat-1.0"
local MINOR_VERSION = tonumber(("$Revision: 38033 $"):match("%d+"))

local useParser30 = AceLibrary:HasInstance("Parser-3.0")

local abilityName, damageType, recipientId
if useParser30 then
	abilityName = "abilityName"
	damageType = "damageType"
	recipientId = "recipientID"
else
	abilityName = "skill"
	damageType = "element"
	recipientId = "victim"
end

if _G.ThreatLib_OUT_OF_DATE or not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then
	_G.ThreatLib_OUT_OF_DATE = true
	return
end

if MINOR_VERSION > _G.ThreatLib_MINOR_VERSION then
	_G.ThreatLib_MINOR_VERSION = MINOR_VERSION
end

local Aura = AceLibrary("SpecialEvents-Aura-2.0")
local ThreatLib = _G.ThreatLib

local ThreatClassModuleCore
if useParser30 then
	ThreatClassModuleCore = ThreatLib:NewModule("ClassCore", "AceDebug-2.0", "AceEvent-2.0", "AceModuleCore-2.0", "AceConsole-2.0", "Parser-3.0")
	ThreatClassModuleCore:SetModuleMixins("AceDebug-2.0", "AceEvent-2.0", "AceConsole-2.0", "Parser-3.0")
else
	ThreatClassModuleCore = ThreatLib:NewModule("ClassCore", "AceDebug-2.0", "AceEvent-2.0", "AceModuleCore-2.0", "AceConsole-2.0")
	ThreatClassModuleCore:SetModuleMixins("AceDebug-2.0", "AceEvent-2.0", "AceConsole-2.0")
end

local new, del, newHash, newSet = ThreatLib.new, ThreatLib.del, ThreatLib.newHash, ThreatLib.newSet

function ThreatClassModuleCore:OnInitialize()
	self:RegisterEvent("AceEvent_FullyInitialized")
end

-- We do this on FullyInitialized because otherwise, talent data is not available on login.
function ThreatClassModuleCore:AceEvent_FullyInitialized()
	self.initialized = true
	local playerClassModule = _G.select(2, UnitClass("player"))
	if self:HasModule(playerClassModule) then
		self:ToggleModuleActive(playerClassModule, true)
		self.activeModule = self:GetModule(playerClassModule)
	end
	if self:HasModule("PET") then
		self:ToggleModuleActive("PET", true)
		self.petModule = self:GetModule("PET")
	end
end

function ThreatClassModuleCore:ClearThreat()
	return self.activeModule:ClearThreat()
end

------------------------------------------
-- Module prototype
------------------------------------------
function ThreatClassModuleCore.modulePrototype:OnEnable()
	if not ThreatClassModuleCore.initialized then
		return
	end 
	self.unitType = "player"

	self.buffThreatMultipliers = 1
	self.debuffThreatMultipliers = 1

	self.buffThreatFlatMods = 0
	self.debuffThreatFlatMods = 0

	self.equipmentThreatModifiers = 1
	self.equipmentBonusThreatModifiers = 1
	self.enchantMods = 1
	self.passiveThreatModifiers = 1
	
	self.redirectThreatTo = nil

	self.itemSets = {}
	self.itemSetsWorn = {}
	
	-- To be filled in by each class module
	self.ClassBuffs = {}
	self.ClassDebuffs = {}
	self.CastHandlers = {}
	self.totalThreatMods = nil
	self.abilityThreatMods = {}
	self.schoolThreatMods = {}
	self.transactions = {}
	
	self.generalThreat = 0
	self.targetThreat = {}

	self:ScanTalents()
	
	self.EXISTING_BUFF = 0
	self.GAINING_BUFF = 1
	self.LOSING_BUFF = 2
	
	self:ClassInit()
	self:RegisterEvent("SpecialEvents_PlayerBuffGained", "buffGained")
	self:RegisterEvent("SpecialEvents_PlayerBuffLost", "buffLost")
	self:RegisterEvent("SpecialEvents_PlayerDebuffGained", "debuffGained")
	self:RegisterEvent("SpecialEvents_PlayerDebuffLost", "debuffLost")
	self:RegisterEvent("SpecialEvents_EquipmentChanged", "equipChanged")
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:RegisterEvent("UNIT_SPELLCAST_SENT")
	self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")
	self:RegisterEvent("CHARACTER_POINTS_CHANGED")

	-- Also need events for pet enter/exit combat
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")

	local sourceID = "player"
	if self.isPetModule then
	 	sourceID = "pet"
	end
	 
	if useParser30 then
		-- Success
		self:RegisterParserEvent({
			eventType = 'Damage',
			sourceID = sourceID,
		}, "parseDamage")
		self:RegisterParserEvent({
			eventType = 'Heal',
			sourceID = sourceID,
		}, "parseHeal")
		self:RegisterParserEvent({
			eventType = 'Gain',
			sourceID = sourceID,
		}, "parseGain")
		self:RegisterParserEvent({
			eventType = 'Cast',
			sourceID = sourceID,
		}, "parseCast")	

		self:RegisterParserEvent({
			eventType = 'Miss',
			sourceID = sourceID,
		}, "parseMiss")

		self:RegisterParserEvent({
			eventType = 'Aura',
			sourceID = sourceID,
		}, "parseAura")
	else
		local events = new(
			'CHAT_MSG_SPELL_PERIODIC_PARTY_BUFFS',
			'CHAT_MSG_SPELL_SELF_BUFF',
			'CHAT_MSG_SPELL_SELF_DAMAGE',
			'CHAT_MSG_SPELL_PERIODIC_SELF_BUFFS',
			'CHAT_MSG_SPELL_CREATURE_VS_SELF_BUFF',
			'CHAT_MSG_COMBAT_SELF_HITS',
			'CHAT_MSG_COMBAT_PET_HITS',
			'CHAT_MSG_SPELL_DAMAGESHIELDS_ON_SELF',
			'CHAT_MSG_COMBAT_CREATURE_VS_SELF_HITS',
			'CHAT_MSG_SPELL_PERIODIC_CREATURE_DAMAGE',
			'CHAT_MSG_SPELL_PERIODIC_SELF_BUFFS'
		)
		local parser = ParserLib:GetInstance("1.1")
		local function parserCombatEvent(info)
			if info.type == "hit" then
				self:parseDamage(info)
			elseif info.type == "heal" then
				self:parseHeal(info)
			elseif info.type == "gain" then
				self:parseGain(info)
			end
		end
		for k,v in pairs(events) do
			parser:RegisterEvent("ThreatClassModule-" .. tostring(self), v, parserCombatEvent)
		end
		events = del(events)
	end
	
	--temp activation of enchant scanning
	self:equipChanged()
	
	self:calcBuffMods()
	self:calcDebuffMods()
end

------------------------------------------------
-- Internal utility methods
------------------------------------------------

function ThreatClassModuleCore.modulePrototype:threatMods()
	if self.totalThreatMods ~= nil then
		return self.totalThreatMods
	end
	self.totalThreatMods = self.buffThreatMultipliers * self.passiveThreatModifiers * self.debuffThreatMultipliers * self.equipmentThreatModifiers * self.equipmentBonusThreatModifiers * self.enchantMods
	return self.totalThreatMods
end

------------------------------------------------
-- Equipment handling
------------------------------------------------
function ThreatClassModuleCore.modulePrototype:getWornSetPieces(name)
	if self.unitType ~= "player" then
		return 0
	end
	if self.itemSetsWorn[name] then
		return self.itemSetsWorn[name]
	end
	
	local ct = 0
	if self.itemSets[name] then
		for i = 1,12 do
			local itemlink = GetInventoryItemLink("player", i)
			if itemlink then
				for k,v in ipairs(self.itemSets[name]) do
					if itemlink:find(v, 1, 1) then
						ct = ct + 1
					end
				end
			end
		end
	end
	self.itemSetsWorn[name] = ct
	return ct
end

function ThreatClassModuleCore.modulePrototype:equipChanged(index, link, oldLink)
	if self.unitType ~= "player" then
		return
	end
	for k,v in pairs(self.itemSetsWorn) do
		self.itemSetsWorn[k] = nil
	end
	--self.totalThreatMods = nil
	
	self.enchantMods = 1
	
	local enchant = (GetInventoryItemLink("player", 15) or ""):match(".-item:%d+:(%d+).*")
	
	-- -2% threat on cloack
	if tonumber(enchant) == 2621 then
		self.enchantMods = self.enchantMods - 0.02
	end
	enchant = (GetInventoryItemLink("player", 10) or ""):match(".-item:%d+:(%d+).*")
	
	-- +2% threat on gloves
	if tonumber(enchant) == 2613 then
		self.enchantMods = self.enchantMods + 0.02
	end
	
	-- metagem -2% threat FIXME: need to make sure that it's active
	enchant = (GetItemGem(GetInventoryItemLink("player", 1) or "", 2) or ""):match(".-item:(%d+).*")
	if tonumber(enchant) == 25897 then
		self.enchantMods = self.enchantMods - 0.02
	end
	
	self.totalThreatMods = nil
	-- Scan equip for any modifying items
end

----------------------------------------------------------------------------------
-- Buff modifier handling
----------------------------------------------------------------------------------
function ThreatClassModuleCore.modulePrototype:AddBuffThreatMultiplier(multiplier)
	self.buffThreatMultipliers = self.buffThreatMultipliers * multiplier
	self.totalThreatMods = nil
end

function ThreatClassModuleCore.modulePrototype:AddBuffThreatModifier(amount)
	self.buffThreatFlatMods = self.buffThreatFlatMods + multiplier
	self.totalThreatMods = nil
end

function ThreatClassModuleCore.modulePrototype:buffChange(buffName, applications, texture, rank, action)
	if ThreatLib.BuffModifiers[buffName] then
		ThreatLib.BuffModifiers[buffName](self, action, rank, applications)
	elseif self.ClassBuffs[buffName] then
		self.ClassBuffs[buffName](self, action, rank, applications)
	end
end

function ThreatClassModuleCore.modulePrototype:calcBuffMods(gainedBuffName)
	self.buffThreatMultipliers = 1
	for buffName, index, applications, texture, rank in Aura:BuffIter(self.unitType) do
		local action = self.EXISTING_BUFF
		if gainedBuffName and gainedBuffName == buffName then
			action = self.GAINING_BUFF
		end
		self:buffChange(buffName, applications, texture, rank, 0) 
	end	
end

function ThreatClassModuleCore.modulePrototype:buffGained(buffName, buffIndex, applications, texture, rank, index) 
	self:calcBuffMods(buffName)
end

function ThreatClassModuleCore.modulePrototype:buffLost(buffName, applications, texture, rank)
	self:buffChange(buffName, applications, texture, rank, self.LOSING_BUFF)
	self:calcBuffMods()
end

----------------------------------------------------------------------------------
-- Debuff modifier handling
----------------------------------------------------------------------------------

function ThreatClassModuleCore.modulePrototype:AddDebuffThreatMultiplier(multiplier)
	self.debuffThreatMultipliers = self.debuffThreatMultipliers * multiplier
	self.totalThreatMods = nil
end

function ThreatClassModuleCore.modulePrototype:AddDebuffThreatModifier(amount)
	self.debuffThreatFlatMods = self.debuffThreatFlatMods + multiplier
	self.totalThreatMods = nil
end

function ThreatClassModuleCore.modulePrototype:debuffChange(buffName, applications, texture, rank, action)
	if ThreatLib.DebuffModifiers[buffName] then
		ThreatLib.DebuffModifiers[buffName](self, action, rank, applications)
	elseif self.ClassDebuffs[buffName] then
		self.ClassDebuffs[buffName](self, action, rank, applications)
	end
end

function ThreatClassModuleCore.modulePrototype:calcDebuffMods(gainingDebuffName)
	self.debuffThreatMultipliers = 1
	for buffName, index, applications, texture, rank in Aura:DebuffIter(self.unitType) do
		local action = self.EXISTING_BUFF
		if buffName == gainingDebuffName then
			action = self.GAINING_BUFF
		end
		self:debuffChange(buffName, applications, texture, rank, action) 
	end	
end

function ThreatClassModuleCore.modulePrototype:debuffGained(debuffName, applications, debuffType, texture, rank, index) 
	self:calcDebuffMods(debuffName)
end

function ThreatClassModuleCore.modulePrototype:debuffLost(debuffName, applications, debuffType, texture, rank)
	self:debuffChange(buffName, applications, texture, rank, self.LOSING_BUFF)
	self:calcDebuffMods()
end

---------------------------------------------------------------------------
-- End buffs/debuffs
---------------------------------------------------------------------------

local partyUnits = ThreatLib.partyUnits

function ThreatClassModuleCore.modulePrototype:parseDamage(info)
	local threat = info.amount
	local skill = info[abilityName]
	
	self:commitOldestTransactionFor(skill)
	
	local handler = self.abilityThreatMods[skill]
	if handler then
		threat = handler(self, threat, true, info.isCrit)
	end
	
	local element = info[damageType]
	handler = self.schoolThreatMods[element]
	if handler then
		threat = handler(self, threat, true, info.isCrit)
	end
	threat = threat * self:threatMods()
	
	self:_addTargetThreat(threat, info.recipientName, self.redirectThreatTo)
	if USE_BUGGED_MISDIRECTION then self:RedirectThreatTo(nil) end
end

function ThreatClassModuleCore.modulePrototype:parseHeal(info)
	if not useParser30 and info.source ~= ParserLib_SELF then
		return
	end
	if not ThreatLib:IsUnitInCombat(self.unitType) then
		return
	end
	
	local unitID
	if useParser30 then
		unitID = info.recipientID
	else
		if info.victim == ParserLib_SELF then
			unitID = "player"
		else
			unitID = partyUnits[info.victim]
		end
	end
	
	local threat = info.amount
	if unitID then
		threat = min(UnitHealthMax(unitID) - UnitHealth(unitID), threat)
	end

	local skill = info[abilityName]
	
	if not self:ExemptGain(skill) then
		local handler = self.abilityThreatMods[skill]
		if handler then
			threat = handler(self, threat, false, info.isCrit)
		end
		threat = threat * ThreatLib.ThreatConstants.healing * self:threatMods()
		self:_addThreat(threat)
	end
end

function ThreatClassModuleCore.modulePrototype:parseGain(info)
	if not useParser30 and info.source ~= ParserLib_SELF then
		return
	end
	if not ThreatLib:IsUnitInCombat(self.unitType) then
		return
	end
	
	local maxgain = UnitManaMax(self.unitType) - UnitMana(self.unitType)
	local amount = min(maxgain, info.amount)
	if not self:ExemptGain(info[abilityName]) then
		if info.attribute == "Mana" then
			self:_addThreat(amount * ThreatLib.ThreatConstants.manaGain * self:threatMods())
		elseif info.attribute == "Rage" then
			self:_addThreat(amount * ThreatLib.ThreatConstants.rageGain * self:threatMods())
		elseif info.attribute == "Energy" then
			self:_addThreat(amount * ThreatLib.ThreatConstants.energyGain * self:threatMods())
		end
	end
end

function ThreatClassModuleCore.modulePrototype:parseMiss(info)
	if info.abilityName then
		self:rollbackTransaction(info.abilityName)
	end
end

function ThreatClassModuleCore.modulePrototype:parseCast(info)
	if self.isPetModule then
		self.currentTarget = info.recipientName
		self:RunCastHandlers(nil, info.abilityName, nil)
		self.currentTarget = nil
	else
		self:commitOldestTransactionFor(info[abilityName])		
	end
end

function ThreatClassModuleCore.modulePrototype:parseAura(info)
end

function ThreatClassModuleCore.modulePrototype:RunCastHandlers(arg1, arg2, arg3)
	if self.CastHandlers[arg2] then
		self.CastHandlers[arg2](self, arg3)
	end	
end

function ThreatClassModuleCore.modulePrototype:PLAYER_REGEN_DISABLED()
	ThreatLib:SetUnitInCombat("player", true)
	self:activatePeriodicTransactionCommit()
end

function ThreatClassModuleCore.modulePrototype:PLAYER_REGEN_ENABLED()
	ThreatLib:SetUnitInCombat("player", false)
	ThreatLib:CancelThreatPublish()
	self:ClearThreat()
	self:rollbackAllTransactions()
	self:deactivatePeriodicTransactionCommit()
end


-- These aren't blizzard events, but we'll stay consistent I guess?
function ThreatClassModuleCore.modulePrototype:PET_REGEN_DISABLED()
	ThreatLib:SetUnitInCombat("pet", true)
	self:activatePeriodicTransactionCommit()
end

function ThreatClassModuleCore.modulePrototype:PET_REGEN_ENABLED()
	ThreatLib:SetUnitInCombat("pet", false)
	ThreatLib:CancelThreatPublishPet()
	self:ClearThreat()
	self:rollbackAllTransactions()
	self:deactivatePeriodicTransactionCommit()
end

function ThreatClassModuleCore.modulePrototype:CHARACTER_POINTS_CHANGED()
	self:ScanTalents()
end

------------------------------------------------
-- Public API
------------------------------------------------

function ThreatClassModuleCore.modulePrototype:ClearThreat()
	self.generalThreat = 0
	for k,v in pairs(self.targetThreat) do
		self.targetThreat[k] = nil
	end	
	ThreatLib:_clearThreat(UnitName(self.unitType))
end

	---------------------------------------------
	-- Threat modification interface [Public API]
	---------------------------------------------
	
local function addtransaction(list, func, param)
	local n = #list
	list[n + 1] = func
	list[n + 2] = param
end

function ThreatClassModuleCore.modulePrototype:RedirectThreatTo(target)
	if target and ThreatLib.partyUnits[target] then
		self.redirectThreatTo = target
		return true
	else
		self.redirectThreatTo = nil
		return false
	end
end

-- Specific-target threat
function ThreatClassModuleCore.modulePrototype:AddTargetThreat(threat)
	-- self:Print(string.format("%s, %s, %s", tostring(threat), tostring(self.currentTarget), tostring(self.redirectThreatTo)))
	if self:hasActiveTransaction() then
		addtransaction(self.activeTransaction.ops, "_addTargetThreat", threat)
		self.activeTransaction.redirectThreatTo = self.redirectThreatTo
		if USE_BUGGED_MISDIRECTION then self:RedirectThreatTo(nil) end
	else
		self:_addTargetThreat(threat, self.currentTarget, self.redirectThreatTo)
		if USE_BUGGED_MISDIRECTION then self:RedirectThreatTo(nil) end
	end
end

-- Specific-target threat
function ThreatClassModuleCore.modulePrototype:MultiplyTargetThreat(modifier)
	if self:hasActiveTransaction() then
		addtransaction(self.activeTransaction.ops, "_multiplyTargetThreat", modifier)
	else
		self:_multiplyTargetThreat(modifier, self.currentTarget)
	end
end

-- Global threat, like heals, mana gain, buffs, etc
function ThreatClassModuleCore.modulePrototype:AddThreat(threat)
	if self:hasActiveTransaction() then
		addtransaction(self.activeTransaction.ops, "_addThreat", threat)
	else
		self:_addThreat(threat)
	end
end

-- Global threat, like heals, mana gain, buffs, etc
function ThreatClassModuleCore.modulePrototype:MultiplyThreat(modifier)
	if self:hasActiveTransaction() then
		addtransaction(self.activeTransaction.ops, "_multiplyThreat", modifier)
	else
		self:_multiplyThreat(modifier)
	end
end

-- Set your target's hate to a given value (Taunt effects, Feign Death minus resists)
function ThreatClassModuleCore.modulePrototype:SetTargetThreat(threat)
	if self:hasActiveTransaction() then
		addtransaction(self.activeTransaction.ops, "_setTargetThreat", threat)
	else
		self:_setTargetThreat(threat, self.currentTarget)
	end
end

-- Reduces all threat by a multiplier. Vanish, Soulshatter, Invisibility, some trinkets use this.
function ThreatClassModuleCore.modulePrototype:ReduceAllThreat(multiplier)
	if self:hasActiveTransaction() then
		addtransaction(self.activeTransaction.ops, "_reduceAllThreat", multiplier)
	else
		self:_reduceAllThreat(multiplier)
	end
end

------------------------------------------------
-- Overridable methods
------------------------------------------------

function ThreatClassModuleCore.modulePrototype:ExemptGain(skill)
	return false
end

function ThreatClassModuleCore.modulePrototype:ScanTalents()
	
end

------------------------------------------------
-- Internal threat modification function, transaction-free
------------------------------------------------

function ThreatClassModuleCore.modulePrototype:_addTargetThreat(threat, currentTarget, redirectTo)
	if partyUnits[currentTarget] then
		return
	end
	if redirectTo then
		ThreatLib:SendThreatTo(redirectTo, currentTarget, threat)
	else
		self.targetThreat[currentTarget] = max(0, (self.targetThreat[currentTarget] or 0) + threat)
		self:TriggerEvent("Threat_ThreatUpdated_Internal", currentTarget, self.targetThreat[currentTarget], self.isPetModule)
	end
end

function ThreatClassModuleCore.modulePrototype:_multiplyTargetThreat(modifier, target)
	if partyUnits[target] then
		return
	end
	self.targetThreat[target] = max(0, (self.targetThreat[target] or 0) * modifier)
	self:TriggerEvent("Threat_ThreatUpdated_Internal", target, self.targetThreat[target], self.isPetModule)
end

function ThreatClassModuleCore.modulePrototype:_addThreat(threat)
	-- We need to not floor this because of things like fade
	-- Instead, we should floor in the primary module
	
	self.generalThreat = self.generalThreat + ( threat / ThreatLib:EncounterMobs() )
	self:TriggerEvent("Threat_ThreatUpdated_Internal", "_GLOBAL_", self.generalThreat, self.isPetModule)
end

function ThreatClassModuleCore.modulePrototype:_multiplyThreat(modifier)
	self.generalThreat = max(0, self.generalThreat * modifier)
	self:TriggerEvent("Threat_ThreatUpdated_Internal", "_GLOBAL_", self.generalThreat, self.isPetModule)
end

function ThreatClassModuleCore.modulePrototype:_setTargetThreat(threat, target)
	self.targetThreat[target] = max(0, threat)
	self:TriggerEvent("Threat_ThreatUpdated_Internal", target, self.targetThreat[target], self.isPetModule)
end

function ThreatClassModuleCore.modulePrototype:_reduceAllThreat(multiplier)
	self.generalThreat = self.generalThreat * multiplier
	for k,v in pairs(self.targetThreat) do
		self.targetThreat[k] = v * multiplier
		ThreatLib:_setThreat(UnitName(self.unitType), k, self.targetThreat[k])
	end	
	-- This will cause an update of all lists
	self:TriggerEvent("Threat_ThreatUpdated_Internal", "_GLOBAL_", self.generalThreat, self.isPetModule)
end

------------------------------------------------
-- Spell transaction stuff
------------------------------------------------
--[[

Transactioning

Begin transaction on UNIT_SPELLCAST_SENT
Store spell name, rank, source, target

Rollback if UNIT_SPELLCAST_FAILED

Commit if CHAT_SPELL_SELF_DAMAGE matches transaction's spell name

Commit if a new transaction is started for a spell that we already have a transaction for

Commit after 2 seconds

Rollback if dodge/parry/miss/block/immune/resist

]]--

function ThreatClassModuleCore.modulePrototype:UNIT_SPELLCAST_SENT(arg1, arg2, arg3, arg4)
	if arg1 ~= self.unitType then
		return
	end
	self.currentTarget = arg4
	self.activeTransaction = self:startTransaction(arg2, arg3)
end

function ThreatClassModuleCore.modulePrototype:UNIT_SPELLCAST_INTERRUPTED(arg1)
	if arg1 ~= self.unitType then
		return
	end
	local skill = self:hasActiveTransaction()
	if skill then
		self:rollbackTransaction(skill)
	end
end

function ThreatClassModuleCore.modulePrototype:UNIT_SPELLCAST_SUCCEEDED(arg1, arg2, arg3)
	if arg1 ~= self.unitType then
		return
	end
	if self:hasActiveTransaction() then
		self.activeTransaction.sent = true
		self.activeTransaction.time = time()
	end
	self:RunCastHandlers(arg1, arg2, arg3)
end

function ThreatClassModuleCore.modulePrototype:hasActiveTransaction()
	return self.activeTransaction and self.activeTransaction.skill
end

function ThreatClassModuleCore.modulePrototype:commitOldestTransactionFor(skill)
	local t = self.transactions[skill]
	if not t then
		return
	end
	t = tremove(t, 1)
	if not t then
		return
	end
	self:commitTransaction(t)
	del(t.ops)
	del(t)
end

function ThreatClassModuleCore.modulePrototype:commitTransaction(transaction)
	for i = 1, #transaction.ops, 2 do
		local call = transaction.ops[i]
		local threat = transaction.ops[i + 1]
		self[call](self, threat, transaction.currentTarget, transaction.redirectThreatTo)		
	end
end

function ThreatClassModuleCore.modulePrototype:rollbackAllTransactions()
	local transactions = self.transactions
	for skill, transactionList in pairs(transactions) do
		for i, t in ipairs(transactionList) do
			del(t.ops)
			del(t)
			transactionList[i] = nil
		end
		del(transactionList)
		transactions[skill] = nil
	end
end


local function commitPeriodicTransactions(this)
	local self = this.self
	local t = time()
	if t - (self.lastCommit or 0) < 1 then
		return
	end
	self.lastCommit = t
	for skill, transactionList in pairs(self.transactions) do
		local offset = 1
		for idx = 1, #transactionList do
			local transaction = transactionList[offset]
			if transaction and transaction.sent and t - transaction.time > 1 then
				self:commitTransaction(transaction)
				transactionList[offset] = del(transactionList[offset])
				tremove(transactionList, offset)
			else
				offset = offset + 1
			end
		end
	end
end

function ThreatClassModuleCore.modulePrototype:startTransaction(skill, rank)
	local t = self.transactions[skill]
	if not t then
		t = new()
		self.transactions[skill] = t
	end
	local v = newHash(
		"skill", skill,
		"rank", rank,
		"sent", false,
		"currentTarget", self.currentTarget,
		"time", time(),
		"ops", new()
	)
	tinsert(t, v)
	return v
end

-- Shifts the oldest entry off of the list and invalidates it
function ThreatClassModuleCore.modulePrototype:rollbackTransaction(skill)
	local t = self.transactions[skill]
	
	-- I -think- it's valid to not have a transaction for a skill, so I'm going to just short-circuit rather than assert. Should doublecheck though.
	if not t or #t == 0 then
		return
	end
	-- assert(#t > 0, "invalid transaction to rollback")
	local v = tremove(t, 1)
	del(v.ops)
	del(v)
end

function ThreatClassModuleCore.modulePrototype:deactivatePeriodicTransactionCommit()
	local f = self.transactionFrame
	if f then
		f:SetScript("OnUpdate", nil)
	end
end

function ThreatClassModuleCore.modulePrototype:activatePeriodicTransactionCommit()
	local f = self.transactionFrame
	if not f then
		f = _G.CreateFrame("Frame", "ThreatLibSpellTransactionPollFrame", _G.UIParent)
		self.transactionFrame = f
		f.self = self
	end
	f:SetScript("OnUpdate", commitPeriodicTransactions)
end

tinsert(ThreatLib.UpvalueFixers, function(lib) ThreatLib = lib end)


